function input_stream(input) {
    let offset = 0, line = 1, column = 1;
    return {
        next,
        peek,
        setCoordination,
        getCoordination,
        eof,
        croak,
    };
    function next() {
        let ch = input.charAt(offset++);
        if (ch == "\n")
            line++, column = 1;
        else
            column++;
        return ch;
    }
    function setCoordination(coordination) {
        offset = coordination.offset;
        line = coordination.line;
        column = coordination.column;
    }
    function getCoordination() {
        return {
            offset,
            line,
            column
        };
    }
    function peek() {
        return input.charAt(offset);
    }
    function eof() {
        return peek() == "";
    }
    function croak(msg) {
        throw new Error(msg + " (" + line + ":" + column + ")");
    }
}

const locStub = {
    filename: '',
    start: { line: 1, column: 1, offset: 0 },
    end: { line: 1, column: 1, offset: 0 }
};
function createAssignStatement(left, right) {
    return {
        type: "ASSIGN" /* ASSIGN */,
        left: left,
        right: right,
        loc: locStub
    };
}
function createListNode(list) {
    return {
        type: "LIST" /* LIST */,
        value: list,
        loc: {
            start: list[0].loc.start,
            end: list[list.length - 1].loc.end,
            filename: list[0].loc.filename
        }
    };
}

const debug = (function () {
    let isDebug = false, count = 0;
    return () => {
        if (count++ > 20 && isDebug) {
            return true;
        }
        return false;
    };
})();
function is_calculate_op_char(ch) {
    let op_chars = ' + - * / % ';
    return op_chars.indexOf(' ' + ch + ' ') >= 0;
}
function is_punc(ch) {
    return ",;(){}#".indexOf(ch) >= 0; // support expr { #{var}:var }
}
const PRECEDENCE = {
    "=": 1,
    "||": 2,
    "&&": 3,
    "<": 7, ">": 7, "<=": 7, ">=": 7, "==": 7, "!=": 7,
    "+": 10, "-": 10,
    "*": 20, "/": 20, "%": 20,
};
/**
 * fill whitespace between tokens which is removed in lexical analyze
 *
 */
function fillWhitespace(tokens) {
    if (tokens.length <= 1)
        return tokens;
    let list = [], whitespaceToken = addNodeEmptyLocation({
        type: "TEXT" /* TEXT */,
        value: ' '
    }), curIndex = 0, curToken, nextToken;
    while (curIndex < tokens.length - 1) {
        curToken = tokens[curIndex];
        nextToken = tokens[curIndex + 1];
        list.push(curToken);
        if (nextToken.loc.start.offset > curToken.loc.end.offset) {
            list.push(whitespaceToken);
        }
        curIndex++;
    }
    list.push(tokens[curIndex]);
    return list;
}
class Environment {
    constructor(parent) {
        this.vars = Object.create(parent ? parent.vars : null);
        this.parent = parent;
    }
    extend() {
        return new Environment(this);
    }
    // lookup: function (name) {
    //     let scope = this;
    //     while (scope) {
    //         if (Object.prototype.hasOwnProperty.call(scope.vars, name))
    //             return scope;
    //         scope = scope.parent;
    //     }
    // },
    get(name) {
        if (name in this.vars)
            return this.vars[name];
        throw new Error("Undefined variable " + name);
    }
    // set: function (name, value) {
    //     let scope = this.lookup(name);
    //     // let's not allow defining globals from a nested environment
    //     if (!scope && this.parent)
    //         throw new Error("Undefined variable " + name);
    //     return (scope || this).vars[name] = value;
    // },
    def(name, value) {
        return this.vars[name] = value;
    }
}
/**
 * todos: optimize deep_clone
 */
function deepClone(obj) {
    return JSON.parse(JSON.stringify(obj));
}
// Some expressions, e.g. sequence and conditional expressions, are never
// associated with generated code, so their source locations are just a empty.
function addNodeEmptyLocation(node) {
    node.loc = locStub;
    return node;
}
function createEmptyNode() {
    return addNodeEmptyLocation({
        type: "EMPTY" /* EMPTY */,
    });
}
function isEmptyNode(node) {
    return node.type === "EMPTY" /* EMPTY */;
}

/**
 *
 * @param {input_stream processed stream method} input
 * { type: NodeTypes.PUNC, value: "(" }           // punctuation: parens((|)), comma(,), semicolon(;) etc.
 * { type: NodeTypes.TEXT, value: "12px" }
 * { type: NodeTypes.VARIABLE, value: "$height" }      // identifiers
 * { type: NodeTypes.KEYWORD, value: NodeTypes.INCLUDE}    // keywords below
 * { type: NodeTypes.PLACEHOLDER, value: "%str" }      //  % started string contains op char '%'
 * { type: NodeTypes.OPERATOR, value: "!=" }            // + - % * / != ==
 */
function lex(input) {
    // let current = null;
    let keywords = ' @extend @mixin @include @import @if @else @error @each ', comparison_op_chars = '!=><', comparison_op_tokens = ['==', '!=', '>=', '<=', '>', '<'];
    return {
        next,
        peek,
        eliminateWhitespace,
        eof,
        getCoordination: input.getCoordination,
        setCoordination: input.setCoordination,
        croak: input.croak
    };
    function is_comment_char(ch) {
        return "/".indexOf(ch) >= 0;
    }
    function is_id_start(ch) {
        return /[$]/.test(ch);
    }
    function skip_comment() {
        read_while(function (ch) { return ch != "\n"; });
        input.next();
    }
    function is_comparison_op_char(ch) {
        // return "+-*/%=&|<>!".indexOf(ch) >= 0;
        return comparison_op_chars.indexOf(ch) >= 0;
    }
    function is_comparison_op_tokens(str) {
        return comparison_op_tokens.indexOf(str) >= 0;
    }
    function is_assign_char(ch) {
        return ":".indexOf(ch) >= 0;
    }
    function is_base_char(ch) {
        return /[a-z0-9_\.\#\@\%\-"'&\[\]]/i.test(ch);
        // return !is_punc(ch);
    }
    function is_id_char_limit(ch) {
        return is_id_start(ch) || /[a-z0-9_-]/i.test(ch); // sass变量名限制
    }
    function is_keyword(kw) {
        return keywords.indexOf(" " + kw + " ") >= 0;
    }
    function is_keyword_start(ch) {
        return ch === '@';
    }
    function is_placeholder_start(ch) {
        return ch === '%';
    }
    function is_whitespace(ch) {
        return " \t\n".indexOf(ch) >= 0;
    }
    function read_while(predicate) {
        let str = "";
        while (!input.eof() && predicate(input.peek()))
            str += input.next();
        return str;
    }
    function read_keyword() {
        let kw = read_while(is_base_char);
        if (!is_keyword(kw)) {
            return input.croak(`Unknown keyword ${kw}`);
        }
        return {
            type: "KEYWORD" /* KEYWORD */,
            value: kw
        };
    }
    function read_assign_char() {
        return {
            type: "ASSIGN" /* ASSIGN */,
            value: input.next()
        };
    }
    function read_ident() {
        let id = read_while(is_id_char_limit);
        return {
            type: "VARIABLE" /* VARIABLE */,
            value: id.trim()
        };
    }
    function generate_op_token(op) {
        return {
            type: "OPERATOR" /* OPERATOR */,
            value: op
        };
    }
    function read_end(endReg) {
        let str = "";
        let is_ended = ch => endReg.test(ch);
        while (!input.eof()) {
            let ch = input.peek();
            if (is_ended(ch)) {
                break;
            }
            else {
                input.next();
                str += ch;
            }
        }
        return str.trim();
    }
    function read_string() {
        /**
         * '#' end eg:
         * .icon-#{$size} {}
         */
        let str = read_end(/[,;{}():#\s]/);
        return {
            type: "TEXT" /* TEXT */,
            value: str
        };
    }
    function maybe_comment(ch) {
        if (is_comment_char(input.peek())) {
            skip_comment();
            return read_next();
        }
        else {
            return generate_op_token(ch);
        }
    }
    function maybe_placeholder(ch) {
        if (is_base_char(input.peek())) {
            return {
                type: "PLACEHOLDER" /* PLACEHOLDER */,
                value: ch + read_while(is_base_char)
            };
        }
        else {
            return generate_op_token(ch);
        }
    }
    function maybe_calculate_op_token(chStr) {
        if (input.peek() === ' ') {
            return generate_op_token(chStr);
        }
        else {
            return {
                type: "TEXT" /* TEXT */,
                value: chStr + read_string().value
            };
        }
    }
    function maybe_comparison_op_token(chStr) {
        if (is_comparison_op_tokens(chStr)) {
            return generate_op_token(chStr);
        }
        else {
            return {
                type: "TEXT" /* TEXT */,
                value: chStr + read_string().value
            };
        }
    }
    function read_punc(ch) {
        return {
            type: "PUNC" /* PUNC */,
            value: ch
        };
    }
    function maybe_punc(ch) {
        if (input.peek() === '{') {
            return read_punc(ch);
        }
        else {
            return {
                type: "TEXT" /* TEXT */,
                value: ch + read_string().value
            };
        }
    }
    function read_next() {
        read_while(is_whitespace);
        if (input.eof())
            return null;
        let ch = input.peek();
        /**
         * comment //  contains op /
         * Todos:还未支持多行注释如：\/**\/
        */
        if (is_comment_char(ch))
            return maybe_comment(input.next());
        if (is_placeholder_start(ch))
            return maybe_placeholder(input.next()); // @extend %message-shared;  contains op '%'
        if (is_assign_char(ch))
            return read_assign_char();
        if (is_id_start(ch))
            return read_ident();
        if (is_punc(ch)) {
            if (ch === '#') {
                return maybe_punc(input.next());
            }
            return read_punc(input.next());
        }
        if (is_keyword_start(ch))
            return read_keyword(); //eg: @extend .message-shared;
        if (is_calculate_op_char(ch))
            return maybe_calculate_op_token(read_while(is_calculate_op_char));
        if (is_comparison_op_char(ch))
            return maybe_comparison_op_token(read_while(is_comparison_op_char));
        if (is_base_char(ch))
            return read_string();
        input.croak("Can't handle character: " + ch);
    }
    /**
     * reset coordination after peek value
     */
    function peek_next() {
        let coordination = input.getCoordination();
        let tok = read_next();
        input.setCoordination(coordination);
        return tok;
    }
    function peek() {
        return peek_next();
    }
    function next() {
        return read_next();
    }
    function eliminateWhitespace() {
        read_while(is_whitespace);
    }
    function eof() {
        return peek() == null;
    }
}

function defaultOnError(error) {
    throw error;
}
function createCompilerError(code, loc, messages, additionalMessage) {
    const msg = (messages || errorMessages)[code] + (additionalMessage || ``);
    const error = new SyntaxError(String(msg));
    error.code = code;
    error.loc = loc;
    return error;
}
const errorMessages = {
    // parse errors
    [0 /* UNKNONWN_TOKEN_TYPE */]: 'Unknown token type.',
    [1 /* INVALID_LOC_POSITION */]: 'Incorrect sourceLocation. start loc should be smaller than end loc',
    [2 /* EXPECT_TEXT_NODE_AFTER_OPERATOR_NODE */]: 'Expect text node after operator node',
};

/**
 *
 * @param {lex processed stream method} input
 */
function parse(input, options) {
    const { filename, source } = options;
    function emitError(code, loc = {
        start: input.getCoordination(),
        end: input.getCoordination(),
        filename
    }) {
        defaultOnError(createCompilerError(code, loc));
    }
    /**
     * end with ';' , eg:
     *
     * $boder : 1px solid red;
     *
     * end with ',' | ')' , eg:
     *
     * @mixin test($param1:1,$param2:2){} // assign expression in @mixin or $include
      */
    function default_right_end_condition() {
        return is_punc(';');
    }
    function set_call_params_args_assign_right_end_condition() {
        assign_right_end_condition = () => is_punc(',') || is_punc(')');
    }
    function reset_assign_right_end_condition() {
        assign_right_end_condition = default_right_end_condition;
    }
    function is_punc(ch) {
        let tok = input.peek();
        return tok && tok.type == "PUNC" /* PUNC */ && (!ch || tok.value == ch) && tok;
    }
    function is_kw(kw) {
        let tok = input.peek();
        return tok && tok.type == "KEYWORD" /* KEYWORD */ && (!kw || tok.value == kw) && tok;
    }
    function is_op(op) {
        let tok = input.peek();
        return tok && tok.type == "OPERATOR" /* OPERATOR */ && (!op || tok.value == op) && tok;
    }
    function is_assign() {
        let tok = input.peek();
        return tok && tok.type === "ASSIGN" /* ASSIGN */;
    }
    function skip_punc(ch) {
        if (is_punc(ch))
            input.next();
        // Todos: no mandatory skip for now
        // else {
        //     input.croak("Expecting punctuation: \"" + ch + "\"");
        // }
    }
    // todos: replace typescript any
    function delimited(start, stop, separator, parser) {
        let a = [], first = true;
        skip_punc(start);
        while (!input.eof()) {
            if (debug())
                break;
            if (is_punc(stop))
                break;
            if (first)
                first = false;
            else
                skip_punc(separator);
            if (is_punc(stop))
                break;
            a.push(parser());
        }
        skip_punc(stop);
        return a;
    }
    let assign_right_end_condition = default_right_end_condition;
    // type parseFunctionType = <T extends Node>(...args:any[]) => T
    function injectLoc(parseFn) {
        return function injectLocInner(...args) {
            input.eliminateWhitespace();
            let start = input.getCoordination(), ast = parseFn.apply(null, args), end = input.getCoordination();
            /**
             *  to resolve stacked injectLoc executed which will lead to  start equal end position
             * */
            // if(ast.loc && ast.loc!==locStub && ast.loc.end.offset>0){
            //     return ast;
            // }
            /**
             * patch ASSIGN , make left node start as start position
             * patch BINARY node position, make binary left most node start as binary node start position
             * */
            if (ast.type === "BINARY" /* BINARY */ || ast.type === "ASSIGN" /* ASSIGN */) {
                let left = args[0];
                while (left.type === "BINARY" /* BINARY */) {
                    left = left.left;
                }
                start = left.loc.start;
            }
            let loc = {
                start,
                end,
                filename
            };
            if (start.offset >= end.offset) {
                emitError(1 /* INVALID_LOC_POSITION */);
            }
            return {
                ...ast,
                loc,
            };
        };
    }
    const parse_error = injectLoc(function parse_error() {
        function parse_list(endCheck = () => !default_right_end_condition()) {
            let list = [];
            while (endCheck()) {
                list.push(dispatchParser());
            }
            return {
                type: "LIST" /* LIST */,
                value: list
            };
        }
        return {
            type: "ERROR" /* ERROR */,
            value: parse_list()
        };
    });
    const consumeNextTokenWithLoc = injectLoc(() => input.next());
    function dispatchParser() {
        if (is_kw('@extend')) {
            input.next();
            return parse_extend();
        }
        if (is_kw('@mixin')) {
            input.next();
            return parse_mixin();
        }
        if (is_kw('@include')) {
            input.next();
            return parse_include();
        }
        if (is_kw('@import')) {
            input.next();
            return parse_import();
        }
        if (is_kw('@if')) {
            input.next();
            return parse_if();
        }
        if (is_kw('@each')) {
            input.next();
            return parse_each();
        }
        if (is_kw('@error')) {
            input.next();
            return parse_error();
        }
        let tok = input.peek();
        if (tok.type === "VARIABLE" /* VARIABLE */ || tok.type === "PLACEHOLDER" /* PLACEHOLDER */) {
            return consumeNextTokenWithLoc();
        }
        if (tok.type === "PUNC" /* PUNC */) {
            if (tok.value === "#") {
                return maybe_key_var_wrapper();
            }
            return consumeNextTokenWithLoc();
        }
        if (tok.type === "TEXT" /* TEXT */) {
            return parse_consecutive_str();
        }
        emitError(0 /* UNKNONWN_TOKEN_TYPE */);
    }
    function parseAssign(left) {
        input.next();
        function parse_assign_right() {
            let right = [];
            while (!assign_right_end_condition()) {
                if (debug())
                    break;
                let result = maybe_binary(dispatchParser(), 0);
                right.push(result);
            }
            return right;
        }
        let right = createListNode(parse_assign_right());
        return createAssignStatement(left, right);
    }
    function maybe_binary(left, left_prec) {
        let tok = is_op();
        if (tok) {
            if (PRECEDENCE[tok.value] > left_prec) {
                tok = consumeNextTokenWithLoc(); //skip op , add loc
                let nextNode = consumeNextTokenWithLoc();
                if (nextNode.type !== "TEXT" /* TEXT */) {
                    emitError(2 /* EXPECT_TEXT_NODE_AFTER_OPERATOR_NODE */, nextNode.loc);
                }
                return maybe_binary({
                    type: "BINARY" /* BINARY */,
                    operator: tok,
                    left: left,
                    right: maybe_binary(nextNode, PRECEDENCE[tok.value]),
                    loc: {
                        start: left.loc.start,
                        end: nextNode.loc.end,
                        filename
                    }
                }, left_prec);
            }
        }
        return left;
    }
    function parse_child(selector) {
        let children = delimited("{", "}", ";", parse_statement);
        return {
            type: "CHILD" /* CHILD */,
            selector,
            children,
            loc: locStub
        };
    }
    function maybe_assign(exp) {
        let expr = exp();
        if (is_assign()) {
            return parseAssign(expr);
        }
        /**
         * handle selector may contain key_var, which should be resolved to list ast
         * .icon-#{$size} {}
        */
        if (is_punc('#')) {
            return maybe_assign(() => {
                return createListNode(expr.type === "LIST" /* LIST */ ?
                    expr.value.concat(dispatchParser())
                    : [expr].concat(dispatchParser()));
            });
        }
        if (is_punc('{')) {
            return parse_child(expr); //passin selector
        }
        return expr;
    }
    function parse_extend() {
        return {
            type: "EXTEND" /* EXTEND */,
            param: consumeNextTokenWithLoc(),
            loc: locStub
        };
    }
    function parse_block_statement() {
        let children = delimited("{", "}", ";", parse_statement);
        return {
            type: "BODY" /* BODY */,
            children,
            loc: locStub
        };
    }
    /**
     * Todos: add @content kw and include body
     */
    function parse_mixin() {
        let id = {
            type: "IDENTIFIER" /* IDENTIFIER */,
            value: input.next().value,
            loc: locStub
        }, params = [];
        if (!is_punc('{')) {
            /**
             * Support default params, which contains assign ':' symbol; which will be processed in parseAssign
             * @mixin replace-text($image,$x:default1, $y:default2) {
              */
            set_call_params_args_assign_right_end_condition();
            params = delimited('(', ')', ',', parse_statement);
        }
        reset_assign_right_end_condition();
        return {
            type: "MIXIN" /* MIXIN */,
            id,
            params,
            body: parse_block_statement(),
            loc: locStub
        };
    }
    function parse_include() {
        let id = {
            type: "IDENTIFIER" /* IDENTIFIER */,
            value: input.next().value,
            loc: locStub
        }, args = []; // @include mixin1;
        if (!is_punc(';')) {
            set_call_params_args_assign_right_end_condition();
            args = delimited('(', ')', ',', parse_statement); // extend like expr or call expr
        }
        reset_assign_right_end_condition();
        return {
            type: "INCLUDE" /* INCLUDE */,
            id,
            args,
            loc: locStub
        };
    }
    function parse_import() {
        function processFilenameExp(exp) {
            exp.value = exp.value.match(/^['"](.+)['"]$/)[1];
            return exp;
        }
        let delimitor = input.peek(), params = [];
        if (!delimitor.value.startsWith("'") && !delimitor.value.startsWith('"')) {
            input.croak(`@import expected with ' or " but encountered ${delimitor}`);
        }
        while (!is_punc(';')) { // @import 'foundation/code', 'foundation/lists';
            params.push(processFilenameExp(consumeNextTokenWithLoc()));
            if (is_punc(',')) {
                skip_punc(',');
            }
        }
        return {
            type: "IMPORT" /* IMPORT */,
            params,
            loc: locStub
        };
    }
    function parse_if() {
        let alternate = null, testExpression, blockStatement;
        testExpression = maybe_binary(dispatchParser(), 0);
        if (!is_punc('{')) {
            input.croak(`@if expect '{' but encountered '${consumeNextTokenWithLoc().value}'`);
        }
        blockStatement = parse_block_statement();
        if (is_kw('@else')) {
            input.next();
            let predictToken = input.peek();
            /**
             * check if it's a @else if  statement
              */
            if (predictToken.type === "TEXT" /* TEXT */ && predictToken.value === 'if') {
                input.next();
                alternate = parse_if();
            }
            else {
                alternate = parse_block_statement();
            }
        }
        return {
            type: "IFSTATEMENT" /* IFSTATEMENT */,
            test: testExpression,
            consequent: blockStatement,
            alternate: alternate,
            loc: locStub
        };
    }
    function parse_each() {
        let left = dispatchParser();
        /**
         * skip "in" expression
          */
        input.next();
        let right = dispatchParser();
        skip_punc('{');
        let blockStatement = parse_statement();
        skip_punc('}');
        return {
            type: "EACHSTATEMENT" /* EACHSTATEMENT */,
            left,
            right,
            body: blockStatement,
            loc: locStub
        };
    }
    function parse_consecutive_str() {
        function maybe_call(exp) {
            let expr = exp();
            if (is_punc('(')) {
                return {
                    ...expr,
                    type: "TEXT" /* TEXT */,
                    value: expr.value + '(' + delimited('(', ')', ',', parse_statement).map(expr => expr.value).join(',') + ')'
                };
            }
            return expr;
        }
        function read_while(predicate) {
            let tokens = [];
            while (!input.eof() && predicate(input.peek()))
                //to resolve test skew(20deg) rotate(20deg);
                tokens.push(maybe_call(consumeNextTokenWithLoc));
            return tokens;
        }
        let list = read_while(tok => tok.type === "TEXT" /* TEXT */), listNode = createListNode(list);
        // no nested ListNode for now, so we flatten and return TextNode
        return {
            type: "TEXT" /* TEXT */,
            value: fillWhitespace(list).map(tok => tok.value).join(''),
            loc: listNode.loc
        };
    }
    /**
     *
     * #{var}
     */
    function maybe_key_var_wrapper() {
        function parse_key_var_wrapper(varKeyStartLoc) {
            skip_punc('{');
            let node = consumeNextTokenWithLoc();
            if (node.type !== "VARIABLE" /* VARIABLE */) {
                input.croak(`${node} should be a variable which starts with '$'`);
            }
            skip_punc('}');
            return {
                ...node,
                type: "VAR_KEY" /* VAR_KEY */,
                loc: {
                    start: varKeyStartLoc,
                    end: node.loc.end,
                    filename
                }
            };
        }
        let token = consumeNextTokenWithLoc();
        if (is_punc('{')) {
            return parse_key_var_wrapper(token.loc.start);
        }
        /**
         * color: #1212; or #selector{}
          */
        let nextToken = consumeNextTokenWithLoc();
        if (nextToken.type !== "TEXT" /* TEXT */) {
            input.croak(`[maybe_key_var_wrapper]: expect str token but received ${nextToken.value}`);
        }
        return {
            type: "TEXT" /* TEXT */,
            value: '#' + nextToken.value,
            loc: {
                start: token.start,
                end: nextToken.end,
                filename
            }
        };
    }
    function parse_statement() {
        return maybe_assign(function () {
            return dispatchParser();
        });
    }
    function parse_prog() {
        let children = [];
        while (!input.eof()) {
            let result = parse_statement();
            children.push(result);
            if (!input.eof())
                skip_punc(";");
        }
        return {
            type: "RootNode" /* RootNode */,
            fileSourceMap: {
                [filename]: source
            },
            children,
            loc: locStub
        };
    }
    return parse_prog();
}

// import runtime_schema_check from './runtime_ast_check'
function baseParse(scss, options) {
    return parse(lex(input_stream(scss)), options);
}

function createCodegenContext(ast, { sourceMap = false }) {
    const context = {
        code: '',
        sourceMap,
        column: 1,
        line: 1,
        offset: 0,
        indentLevel: 0,
        // source: ast.source,
        push(code, node) {
            context.code += code;
        },
        indent() {
            newline(++context.indentLevel);
        },
        deindent(withoutNewLine = false) {
            if (withoutNewLine) {
                --context.indentLevel;
            }
            else {
                newline(--context.indentLevel);
            }
        },
        newline() {
            newline(context.indentLevel);
        }
    };
    function newline(n) {
        context.push('\n' + `  `.repeat(n));
    }
    return context;
}
function generate(ast, options = {}) {
    const context = createCodegenContext(ast, options);
    ast.children.forEach((node) => genNode(node, context));
    return {
        ast,
        code: context.code,
        // SourceMapGenerator does have toJSON() method but it's not in the types
        map: context.map ? context.map.toJSON() : undefined
    };
}
function genNode(node, context) {
    if (!node) {
        console.error('node', node, 'context', context);
        return;
    }
    switch (node.type) {
        case "TEXT" /* TEXT */:
            genText(node, context);
            break;
        case "ASSIGN" /* ASSIGN */:
            genAssign(node, context);
            break;
        case "CHILD" /* CHILD */:
            genChild(node, context);
            break;
        case "EMPTY" /* EMPTY */:
            break;
        default:
            throw new Error("Don't know how to genNode for " + JSON.stringify(node));
    }
}
function genText(node, context) {
    context.push(node.value, node);
}
function genAssign(node, context) {
    const { push } = context;
    genNode(node.left, context);
    push(':');
    genNode(node.right, context);
    push(';');
}
function genChild(node, context) {
    const { push, deindent, indent, newline } = context;
    genNode(node.selector, context);
    push('{');
    indent();
    node.children.forEach((node, index) => {
        if (index && node.type !== "EMPTY" /* EMPTY */) {
            newline();
        }
        genNode(node, context);
    });
    deindent();
    push('}');
    newline();
}

function tranform_nest(ast) {
    function flatten_child(child, arr = []) {
        function flatten(child, parentSelector = '') {
            /**
             * https://sass-lang.com/documentation/style-rules/parent-selector
             * support parent selector with loosening restriction
             */
            let containParentSelector = false;
            let selector = child.selector;
            if (selector.value.indexOf('&') >= 0) {
                containParentSelector = true;
                selector.value = selector.value.replace(/&/, parentSelector);
            }
            if (!containParentSelector) {
                selector.value = (parentSelector + ' ' + selector.value).trim();
            }
            /**
             * 清理空的选择器
              */
            child.selector.value && arr.push(child);
            /**
             *
              */
            child.children.forEach((exp, index) => {
                if (exp.type === "CHILD" /* CHILD */) {
                    child.children.splice(index, 1, createEmptyNode());
                    flatten(exp, selector.value);
                }
            });
            return arr;
        }
        return flatten(child);
    }
    function toplevel(ast) {
        let children = [];
        ast.children.forEach(exp => {
            if (exp.type === "CHILD" /* CHILD */) {
                children = children.concat(...flatten_child(exp));
            }
            else {
                children.push(exp);
            }
        });
        ast.children = children;
        return ast;
    }
    return toplevel(ast);
}

function transform_extend(ast) {
    /**
     * transform
     *
     * .basic{}
     * .primary{
     *     @extend .basic;
     * }
     *
     * to
     *
     * {'.basic':['primary']}
     */
    let extendSelectorPair = {};
    function is_placeholder(exp) {
        return exp.type === "PLACEHOLDER" /* PLACEHOLDER */;
    }
    function rm_empty_child(exp) {
        return exp.type === "CHILD" /* CHILD */ && (is_placeholder(exp.selector) || exp.children.length === 0) ? null : exp;
    }
    function collect_extend(exp) {
        function collect(child) {
            child.children = child.children.map(exp => {
                if (exp.type === "EXTEND" /* EXTEND */) {
                    extendSelectorPair[exp.param.value] = (extendSelectorPair[exp.param.value] || []).concat(child.selector.value);
                    return null;
                }
                return exp;
            }).filter(exp => exp !== null);
            return child;
        }
        return exp.type === "CHILD" /* CHILD */ ? collect(exp) : exp;
    }
    function transform_extend(exp) {
        function transform(child) {
            if (is_placeholder(child.selector)) {
                child.selector = {
                    type: "TEXT" /* TEXT */,
                    value: extendSelectorPair[child.selector.value].join(','),
                    loc: child.selector.loc
                };
            }
            else {
                child.selector = {
                    type: "TEXT" /* TEXT */,
                    value: extendSelectorPair[child.selector.value].concat(child.selector.value).join(','),
                    loc: child.selector.loc
                };
            }
            return child;
        }
        return exp.type === "CHILD" /* CHILD */ && extendSelectorPair[exp.selector.value] ? transform(exp) : exp;
    }
    function toplevel(ast) {
        ast.children = ast.children.map(exp => collect_extend(exp));
        ast.children = ast.children.map(exp => transform_extend(exp));
        ast.children = ast.children.map(exp => rm_empty_child(exp)).filter(exp => exp !== null);
        return ast;
    }
    return toplevel(ast);
}

let transformMiddleware = [tranform_nest, transform_extend];
var transformChain = (ast) => transformMiddleware.reduce((ast, middleware) => {
    return middleware.call(null, ast);
}, ast);

function createTransformContext(root, { nodeTransforms = [], onError = defaultOnError, sourceDir = './' }) {
    const context = {
        onError,
        nodeTransforms,
        root,
        sourceDir,
        env: new Environment(null),
        parent: null,
        currentNode: root,
        childIndex: 0,
        onNodeRemoved: () => { },
        replaceNode(node) {
            if (!context.parent) {
                throw new Error(`Cannot replace root node.`);
            }
            context.parent.children[context.childIndex] = context.currentNode = node;
            return node;
        }
    };
    return context;
}
function transform(root, options) {
    const context = createTransformContext(root, options);
    traverseNode(root, context);
    // transformChain will be slowly replaced by transform plugins if possible
    root = transformChain(root);
}
function traverseChildren(parent, context) {
    let i = 0;
    const nodeRemoved = () => {
        i--;
    };
    for (; i < parent.children.length; i++) {
        const child = parent.children[i];
        context.parent = parent;
        context.childIndex = i;
        context.onNodeRemoved = nodeRemoved;
        traverseNode(child, context);
    }
    parent.children = parent.children.filter((node) => node.type !== "EMPTY" /* EMPTY */);
}
/*
only traverse top level for now
this sass compiler combined evaluate and compiling to direct target code so it needs to construct executing env while compiling
so it will be difficult for pure transform plugins
different from compile to target code which is evaluatable
  */
function traverseNode(node, context) {
    context.currentNode = node;
    // apply transform plugins
    const { nodeTransforms } = context;
    if (node.type !== "RootNode" /* RootNode */) {
        for (let i = 0; i < nodeTransforms.length; i++) {
            nodeTransforms[i](node, context);
        }
    }
    if (node.type === "RootNode" /* RootNode */) {
        traverseChildren(node, context);
    }
}

function processExpression(node, context) {
    function dispatchExpression(node, context) {
        switch (node.type) {
            /**
             * Expression
             */
            case "TEXT" /* TEXT */: return transformText(node);
            case "PUNC" /* PUNC */: return transformPunc(node);
            case "OPERATOR" /* OPERATOR */: return transformOp(node);
            case "VARIABLE" /* VARIABLE */: return transformVar(node, context);
            case "VAR_KEY" /* VAR_KEY */: return transformVarKey(node, context);
            case "LIST" /* LIST */: return transformList(node, context);
            case "BINARY" /* BINARY */: return transformBinary(node, context);
            default:
                throw new Error("Don't know how to processExpression expression for " + JSON.stringify(node));
        }
    }
    /**
     * Solve situation, assign value with NodeTypes.PUNC, eg:
     * $font: Helvetica, sans-serif;
     */
    function transformPunc(node) {
        return {
            ...node,
            type: "TEXT" /* TEXT */
        };
    }
    /**
     * Solve situation, selector with operators, eg:
     * .a > b{}
     */
    function transformOp(node) {
        return transformPunc(node);
    }
    function transformText(node) {
        return node;
    }
    function transformVar(node, context) {
        return {
            ...node,
            type: "TEXT" /* TEXT */,
            value: context.env.get(node.value)
        };
    }
    function transformVarKey(node, context) {
        return transformVar(node, context);
    }
    function transformList(node, context) {
        return {
            type: "TEXT" /* TEXT */,
            loc: node.loc,
            value: fillWhitespace(node.value).map((item) => {
                return dispatchExpression(item, context).value;
            }).join('').trim()
        };
    }
    /**
     * Todos:add more unit type eg: pt px etc
     *
     */
    function transformBinary(node, context) {
        let hasPercent = false, unitExtracted = false, unit = "";
        function parseFloatFn(str) {
            if (isNaN(parseFloat(str))) {
                return str;
            }
            if (str === "100%") {
                hasPercent = true;
                return 1;
            }
            /**
             * make the first encountered unit as the final unit
              */
            if (!unitExtracted) {
                unitExtracted = true;
                let matched = str.match(/\d+([a-z]*)/);
                unit = matched && matched.length > 0 ? matched[1] : '';
            }
            return parseFloat(str);
        }
        function transformVal(str) {
            function transformPercent(str) {
                return parseFloat(str) * 100 + "%";
            }
            function addUnit(str) {
                return str + unit;
            }
            return hasPercent ? transformPercent(str) : addUnit(str);
        }
        function evaluateBinary(node) {
            const opAcMap = {
                // '=': (left, right) => left = right,
                '||': (left, right) => left || right,
                '&&': (left, right) => left && right,
                '==': (left, right) => left == right,
                '!=': (left, right) => left != right,
                '>=': (left, right) => left >= right,
                '<=': (left, right) => left <= right,
                '+': (left, right) => left + right,
                '-': (left, right) => left - right,
                '/': (left, right) => left / right,
                '*': (left, right) => left * right,
                '%': (left, right) => left % right,
            };
            if (node.type === "TEXT" /* TEXT */)
                return parseFloatFn(node.value);
            if (node.type === "VARIABLE" /* VARIABLE */)
                return evaluateBinary({
                    ...node,
                    type: "TEXT" /* TEXT */,
                    value: context.env.get(node.value),
                });
            if (node.type === "BINARY" /* BINARY */)
                return opAcMap[node.operator.value](evaluateBinary(node.left), evaluateBinary(node.right));
            throw new Error("Don't know how to evaluateBinary type: " + node.type);
        }
        let value = evaluateBinary(node);
        return {
            ...node,
            type: "TEXT" /* TEXT */,
            value: transformVal(value),
        };
    }
    return dispatchExpression(node, context);
}

const processAssign = (node, context) => {
    if (node.left.type === "VARIABLE" /* VARIABLE */) {
        context.env.def(node.left.value, processExpression(node.right, context).value);
        return createEmptyNode();
    }
    if (node.left.type === "VAR_KEY" /* VAR_KEY */) {
        node.left = processExpression(node.left, context);
    }
    node.right = processExpression(node.right, context);
    return node;
};

const transformStatement = (node, context) => {
    processStatement(node, context);
};
function processStatement(node, context) {
    function dispatchStatement(node, context) {
        // console.log(context.env.parent)
        switch (node.type) {
            /**
            * Statement
            */
            case "ASSIGN" /* ASSIGN */: return processAssign(node, context);
            case "MIXIN" /* MIXIN */: return transform_mixin(node, context);
            case "INCLUDE" /* INCLUDE */: return transform_include(node, context);
            case "CHILD" /* CHILD */:
            case "BODY" /* BODY */:
                return transform_child_or_body(node, context);
            case "EXTEND" /* EXTEND */: return node;
            case "IFSTATEMENT" /* IFSTATEMENT */: return transform_if(node, context);
            case "EACHSTATEMENT" /* EACHSTATEMENT */: return transform_each(node, context);
            case "ERROR" /* ERROR */:
                throw new Error(processExpression(node.value, context).value);
            default:
                throw new Error("Don't know how to compile statement for " + JSON.stringify(node));
        }
    }
    /**
     * any expressions separated with spaces or commas count as a list
     */
    function transform_each(node, context) {
        let restoredContext = deepClone(node), right = processExpression(restoredContext.right, context), scope = context.env.extend(), list = [];
        if (right.type === "TEXT" /* TEXT */) {
            list = right.value.split(/,|\s+/).filter(val => val !== '');
        }
        let children = list.map(val => {
            /**
             * restore context with iterate multiple times
             */
            restoredContext = deepClone(node);
            scope.def(restoredContext.left.value, val);
            return dispatchStatement(restoredContext.body, { ...context, env: scope });
        });
        /**
         * return a created an empty selector child whose children will be flattened in transform_nest
         */
        let child = {
            ...node,
            type: "CHILD" /* CHILD */,
            selector: addNodeEmptyLocation({
                type: "TEXT" /* TEXT */,
                value: ''
            }),
            children
        };
        return child;
    }
    /**
     * context will be restored with function
     */
    function transform_if(node, context) {
        function is_true_exp(expression, context) {
            let resultExp = expression;
            /**
             *  1. if $bool { }
             *  2. if true { }
            */
            if (resultExp.type === "VARIABLE" /* VARIABLE */ || resultExp.type === "BINARY" /* BINARY */) {
                resultExp = processExpression(expression, context);
            }
            if (resultExp.value === "0" || resultExp.value === "false") {
                return false;
            }
            return true;
        }
        if (is_true_exp(node.test, context)) {
            return dispatchStatement(node.consequent, context);
        }
        else if (node.alternate) {
            return dispatchStatement(node.alternate, context);
        }
        else {
            return createEmptyNode();
        }
    }
    /**
     * transform @mixin -> set to env -> delete @mixin ast
     */
    function transform_mixin(node, context) {
        function make_function() {
            /**
             * deep clone function statement to restore context when called multiple times
             */
            let restoredContext = deepClone(node), params = restoredContext.params, scope = context.env.extend();
            function handle_params_default_value(params) {
                return params.map((param) => {
                    let ret = param;
                    if (param.type === "ASSIGN" /* ASSIGN */) {
                        ret = {
                            type: "VARIABLE" /* VARIABLE */,
                            value: param.left.value
                        };
                        dispatchStatement(param, { ...context, env: scope });
                    }
                    return ret;
                });
            }
            params = handle_params_default_value(params);
            params.forEach((param, i) => {
                if (param.type === "VARIABLE" /* VARIABLE */ && arguments[i]) {
                    scope.def(param.value, arguments[i]);
                }
            });
            return dispatchStatement(restoredContext.body, { ...context, env: scope });
        }
        context.env.def(node.id.value, make_function);
        return createEmptyNode();
    }
    function transform_include(node, context) {
        let func = context.env.get(node.id.value);
        return func.apply(null, node.args.map(arg => {
            /**
             * @include avatar(100px, $circle: false);
             */
            if (arg.type === "ASSIGN" /* ASSIGN */) {
                return processExpression(arg.right, context).value;
            }
            else {
                return processExpression(arg, context).value;
            }
        }));
    }
    function transform_child_or_body(node, context) {
        function flatten_included_body(children) {
            let arr = [];
            children.forEach(child => {
                if (child.type === "BODY" /* BODY */) {
                    arr = arr.concat(flatten_included_body(child.children));
                }
                else {
                    arr.push(child);
                }
            });
            return arr;
        }
        let scope = context.env.extend();
        node.children = node.children.map(child => dispatchStatement(child, { ...context, env: scope })).filter(node => !isEmptyNode(node));
        if (node.selector && node.selector.type === "LIST" /* LIST */) {
            node.selector = processExpression(node.selector, { ...context, env: scope });
        }
        /**
         * resolve BlockStatement/body
         *
         * @include include_function_body
         * @if which contain BlockStatement
          */
        node.children = flatten_included_body(node.children);
        return node;
    }
    return context.replaceNode(dispatchStatement(node, context));
}

function baseCompile(scss, options = {
    filename: 'default.scss',
    source: ''
}) {
    let ast = baseParse(scss, options);
    transform(ast, {
        nodeTransforms: [transformStatement],
        ...options
    });
    return generate(ast, options);
}

export { baseCompile as compile, baseParse as parse };
//# sourceMappingURL=tiny-sass-compiler.esm-browser.prod.js.map
